@@ -125,7 +125,7 @@ distance_matrix_to_geometry_conversion_notes = """
 
 if bUseNumba :
         @jit(nopython=True)
-        def distance_matrix_to_absolute_coordinates ( D , bSquared = False, n_dimensions=2 ):
+        def distance_matrix_to_absolute_coordinates ( D , bSquared = False, n_dimensions=2 , bLegacy=True ) :
                 # C++ https://github.com/richardtjornhammar/RichTools/commit/be0c4dfa8f61915b0701561e39ca906a9a2e0bae
                 if not bSquared :
                         D = D**2.
@@ -139,10 +139,16 @@ if bUseNumba :
                 U,S,Vt = np.linalg.svd ( D , full_matrices = True )
                 S[DIM:] *= 0.
                 Z = np.diag(S**0.5)[:,:DIM]
-                xr = np.dot( Z.T,Vt )
-                return ( xr )
+                if bLegacy :
+                    print ( "YOU ARE RUNNING THE LEGACY IMPLEMENTATION. THE RETURN MATRIX WILL BE CHANGED TO ITS TRANSPOSE" )
+                    print ( "TO USE THE NEW VERSION SET bLegacy ARGUMENT TO FALSE" )
+                    xr = np.dot( Z.T,Vt )
+                    return ( xr )
+                else:
+                    xr = np.tensordot( Z.T,Vt, axes=(0,1) )
+                    return ( xr.T )
 else :
-        def distance_matrix_to_absolute_coordinates ( D , bSquared = False, n_dimensions=2 ):
+        def distance_matrix_to_absolute_coordinates ( D , bSquared = False, n_dimensions=2 , bLegacy=True ):
                 # C++ https://github.com/richardtjornhammar/RichTools/commit/be0c4dfa8f61915b0701561e39ca906a9a2e0bae
                 if not bSquared :
                         D = D**2.
@@ -156,8 +162,14 @@ else :
                 U,S,Vt = np.linalg.svd ( D , full_matrices = True )
                 S[DIM:] *= 0.
                 Z = np.diag(S**0.5)[:,:DIM]
-                xr = np.dot( Z.T,Vt )
-                return ( xr )
+                if bLegacy :
+                    print ( "YOU ARE RUNNING THE LEGACY IMPLEMENTATION. THE RETURN MATRIX WILL BE CHANGED TO ITS TRANSPOSE" )
+                    print ( "TO USE THE NEW VERSION SET bLegacy ARGUMENT TO FALSE" )
+                    xr = np.dot( Z.T,Vt )
+                    return ( xr )
+                else:
+                    xr = np.tensordot( Z.T,Vt, axes=(0,1) )
+                    return ( xr.T )
 
 if bUseNumba :
         @jit(nopython=True)
@@ -175,7 +187,7 @@ if bUseNumba :
                 #
                 # https://github.com/richardtjornhammar/RichTools/blob/master/src/cluster.cc
                 # ADDED TO RICHTOOLS HERE: https://github.com/richardtjornhammar/RichTools/commit/74b35df9c623bf03570707a24eafe828f461ed90#diff-25a6634263c1b1f6fc4697a04e2b9904ea4b042a89af59dc93ec1f5d44848a26
-                # CONNECTIVITY SEARCH FOR (connectivity) CONNECTIVITY 
+                # CONNECTIVITY SEARCH FOR (connectivity) CONNECTIVITY
                 #
                 nr_sq,mr_sq = np.shape(B)
                 if nr_sq != mr_sq :
@@ -346,6 +358,226 @@ else :
                     L .append( ids )
             return ( L )
 
+if bUseNumba :
+        @jit(nopython=True)
+        def connectivity_boolean ( B , bVerbose=False ) :
+                description = """ This is a cutoff based clustering algorithm. The intended use is to supply a distance matrix and a cutoff value (then becom>
+m. It has been employed for statistical analysis work as well as the original application where it was employed to segment molecular systems."""
+                if bVerbose :
+                        print ( "CONNECTIVITY CLUSTERING OF ", np.shape(B), " MATRIX" )
+                # PYTHON ADAPTATION OF MY C++ CODE THAT CAN BE FOUND IN
+                # https://github.com/richardtjornhammar/RichTools/blob/master/src/cluster.cc
+                # AROUND LINE 2277
+                # CONSIDER COMPILING AND USING THAT AS A MODULE INSTEAD OF THIS SINCE IT IS
+                # A LOT FASTER
+                # FOR A DESCRIPTION READ PAGE 30 (16 INTERNAL NUMBERING) of:
+                # https://kth.diva-portal.org/smash/get/diva2:748464/FULLTEXT01.pdf
+                #
+                # https://github.com/richardtjornhammar/RichTools/blob/master/src/cluster.cc
+                # ADDED TO RICHTOOLS HERE: https://github.com/richardtjornhammar/RichTools/commit/74b35df9c623bf03570707a24eafe828f461ed90#diff-25a6634263c1b>
+                # CONNECTIVITY SEARCH FOR (connectivity) CONNECTIVITY
+                #
+                nr_sq,mr_sq = np.shape(B)
+                if nr_sq != mr_sq :
+                        print ( 'ERROR: FAILED' )
+                N = mr_sq
+                res , nvisi, s, NN, ndx, C = [0], [0], [0], [0], [0], 0
+                res .append(0)
+                for i in range(N) :
+                        nvisi.append(i+1)
+                        res.append(0); res.append(0)
+                        ndx.append(i)
+
+                res   = res[1:]
+                nvisi = nvisi[1:]
+                ndx   = ndx[1:]
+                while ( len(ndx)>0 ) :
+                        i = ndx[-1] ; ndx = ndx[:-1]
+                        NN = []
+                        if ( nvisi[i]>0 ) :
+                                C-=1
+                                for j in range(N) :
+                                        if ( B[i,j] ) :
+                                                NN.append(j)
+                                while ( len(NN)>0 ) :
+                                        # back pop_back
+                                        k = NN[-1]; NN = NN[:-1]
+                                        nvisi[k] = C
+                                        for j in range(N):
+                                                if ( B[j,k] ) :
+                                                        for q in range(N) :
+                                                                if ( nvisi[q] == j+1 ) :
+                                                                        NN.append(q)
+                if bVerbose : # VERBOSE
+                        print ( "INFO "+str(-1*C) +" clusters" )
+                Nc = [ 0 for i in range(-1*C) ]
+                for q in range(N) :
+                        res[  q*2+1 ] = q;
+                        res[  q*2   ] = nvisi[q]-C;
+                        Nc [res[q*2]]+= 1;
+                        if bVerbose :
+                                print ( " "+str(res[q*2])+" "+str(res[2*q+1]) )
+                if bVerbose:
+                        for i in range(-1*C) :
+                                print( "CLUSTER "  +str(i)+ " HAS " + str(Nc[i]) + " ELEMENTS")
+                return ( Nc , np.array(res[:-1]).reshape(-1,2) )
+else:
+        def connectivity_boolean ( B , bVerbose=False ) :
+                description = """ This is a cutoff based clustering algorithm. The intended use is to supply a distance matrix and a cutoff value (then becom>
+m. It has been employed for statistical analysis work as well as the original application where it was employed to segment molecular systems."""
+                if bVerbose :
+                        print ( "CONNECTIVITY CLUSTERING OF ", np.shape(B), " MATRIX" )
+                # PYTHON ADAPTATION OF MY C++ CODE THAT CAN BE FOUND IN
+                # https://github.com/richardtjornhammar/RichTools/blob/master/src/cluster.cc
+                # AROUND LINE 2277
+                # CONSIDER COMPILING AND USING THAT AS A MODULE INSTEAD OF THIS SINCE IT IS
+                # A LOT FASTER
+                # FOR A DESCRIPTION READ PAGE 30 (16 INTERNAL NUMBERING) of:
+                # https://kth.diva-portal.org/smash/get/diva2:748464/FULLTEXT01.pdf
+                #
+                # https://github.com/richardtjornhammar/RichTools/blob/master/src/cluster.cc
+                # ADDED TO RICHTOOLS HERE: https://github.com/richardtjornhammar/RichTools/commit/74b35df9c623bf03570707a24eafe828f461ed90#diff-25a6634263c1b>
+                # CONNECTIVITY SEARCH FOR (connectivity) CONNECTIVITY
+                #
+                nr_sq,mr_sq = np.shape(B)
+                if nr_sq != mr_sq :
+                        print ( 'ERROR: FAILED' )
+                N = mr_sq
+                res , nvisi, s, NN, ndx, C = [0], [0], [0], [0], [0], 0
+                res .append(0)
+                for i in range(N) :
+                        nvisi.append(i+1)
+                        res.append(0); res.append(0)
+                        ndx.append(i)
+
+                res   = res[1:]
+                nvisi = nvisi[1:]
+                ndx   = ndx[1:]
+                while ( len(ndx)>0 ) :
+                        i = ndx[-1] ; ndx = ndx[:-1]
+                        NN = []
+                        if ( nvisi[i]>0 ) :
+                                C-=1
+                                for j in range(N) :
+                                        if ( B[i,j] ) :
+                                                NN.append(j)
+                                while ( len(NN)>0 ) :
+                                        # back pop_back
+                                        k = NN[-1]; NN = NN[:-1]
+                                        nvisi[k] = C
+                                        for j in range(N):
+                                                if ( B[j,k] ) :
+                                                        for q in range(N) :
+                                                                if ( nvisi[q] == j+1 ) :
+                                                                        NN.append(q)
+                if bVerbose : # VERBOSE
+                        print ( "INFO "+str(-1*C) +" clusters" )
+                Nc = [ 0 for i in range(-1*C) ]
+                for q in range(N) :
+                        res[  q*2+1 ] = q;
+                        res[  q*2   ] = nvisi[q]-C;
+                        Nc [res[q*2]]+= 1;
+                        if bVerbose :
+                                print ( " "+str(res[q*2])+" "+str(res[2*q+1]) )
+                if bVerbose:
+                        for i in range(-1*C) :
+                                print( "CLUSTER "  +str(i)+ " HAS " + str(Nc[i]) + " ELEMENTS")
+                return ( Nc , np.array(res[:-1]).reshape(-1,2) )
+
+
+def clustering_bounds ( distm:np.array ) -> tuple :
+    #
+    # RETURN THE BOUNDS FOR WHEN THE CLUSTERING SOLUTION IS
+    # DISCONNECTED CONTRA FULLY CONNECTED
+    #
+    d = distm.copy()
+    ma = np.min(np.max(d,0))
+    for i in range(len(d)):
+        d[i,i] = np.inf
+    return ( np.min(d) , ma )
+
+
+def pidx2lidx ( n:int,i:int,j:int )->int :
+    from scipy.special import binom
+    p = i if i>j else j
+    i = i if i<j else j
+    j = p
+    return ( int ( binom(n,2) - binom(n-i,2) + ( j-i-1 ) )  )
+
+def linear2coord ( N:int, bType=0 ) -> dict :
+    if bType == 0 :
+        return ( { pidx2lidx(N,i,j):(i,j)  for i in range(N) for j in range(N) if j>i } )
+    from scipy.spatial.distance import squareform
+    tmpsq =  squareform ( [ i for i in range(int( N*(N-1)*0.5 )) ] )
+    return (  { tmpsq[i,j]:(i,j) for i in range(N) for j in range(N) if j>i } )
+
+if True :
+    #@njit #@jit( nopython=True )
+    def link_clustering_core ( pdp:np.array , N:int ) :
+        pdp  = sorted(pdp)[::-1]
+        idxs = [ tuple([n]) for n in range(N) ]
+        ids0 = idxs.copy()
+        cluster_ledger = [ [ 0, idxs ] ]
+        for i in range(len(pdp)) :
+            p_ = pdp.pop()
+            m_ = set( p_[1] )
+            for j_ in range(len(idxs)) :
+                s_ = idxs.pop()
+                if len(m_&set(s_))>0:
+                    m_ = m_|set(s_)
+                else :
+                    idxs = [ s_,*idxs ]
+            idxs = [ tuple(sorted(m_)) ,*idxs ]
+            if set(idxs) == set(ids0) :
+                continue
+            cluster_ledger.append( [ p_[0] , idxs.copy() ] )
+            ids0 = idxs
+            if len(idxs) == 1 : # WE ARE FULLY CONNECTED
+                break
+        return ( cluster_ledger )
+
+def link_clustering ( input:np.array ) :
+    from scipy.spatial.distance import pdist,squareform
+    import time,gc
+    #
+    # INPUT CHECK
+    #
+    nm = np.shape(input)
+    if len(nm) == 1 : # ASSUME PDIST ARRAY
+        pdc = input.copy()
+        distm = squareform( pdc )
+    else :
+        if nm[0] == nm[1] : # ASSUME DISTANCE MATRIX
+            distm = input.copy()
+            pdc = squareform( input )
+        else : # ASSUME COORDINATES
+            pdc     = pdist ( input )
+            distm = squareform ( pdc )
+    nm = np.shape(distm)
+    N  = nm[0]
+    lookup  = linear2coord( N )
+    fully_connected_at = np.min(np.max( distm ,0 ))
+    del distm
+    gc.collect()
+    pdp     = [ (d,lookup[i] ) for d,i in zip(pdc,range(len(pdc))) if d <= fully_connected_at ]
+    del lookup
+    gc.collect()
+    del pdc
+    gc.collect()
+    results = link_clustering_core( pdp , N )
+    return ( results )
+
+
+def connectivity_clustering( distm:np.array, cutoff:float, bBool:bool=False ) -> tuple :
+    if bBool :
+        # THIS WILL BE FASTER AND MORE MEMORY EFFICIENT
+        # IF THE DISTM IS LARGE
+        distm_b = distm<=cutoff
+        res = connectivity_boolean( distm_b )
+    else :
+        res = connectivity( distm , cutoff )
+    return ( res )
+        
 
 clustering_algorithm = None
 clustering_algorithm = sc.KMeans(10) # CHOOSE SOMETHING YOU LIKE NOT THIS
@@ -701,7 +933,6 @@ def calculate_rdf ( particles_i = None , particles_o = None , nbins=100 ,
                     bRemoveZeros = False )""")
         exit ( 1 )
 
-
 def unpack ( seq ) : # seq:Union -> Union
     if isinstance ( seq,(list,tuple,set)) :
         yield from ( x for y in seq for x in unpack(y) )
@@ -718,95 +949,252 @@ def rem ( a:list , H:list ) -> list :
         a .pop(hp)
     return(a)
 
-def linkage ( distm:np.array , command:str = 'max' ) -> dict :
+def nppop(A:np.array, irow:int=None, jcol:int=None ) -> list[np.array] :
+    # ASSUMES ROW MAJOR ORDER
+    rrow:np.array() = None
+    rcol:np.array() = None
+    N = len(A)
+    M0,M1 = np.shape(A)
+    if not irow is None :
+        rrow = A[irow,:]
+        A    = np.delete(A,range(N*irow,N*(irow+1))).reshape(-1,N)
+        M0   = M0-1
+    if not jcol is None :
+        rcol = A[:,jcol]
+        A    = np.delete(A,range(jcol,len(A.reshape(-1)),N) )
+        M1   = M1-1
+    return ( [rrow,rcol,A.reshape(M0,M1)] )
+
+def link1_ ( D:np.array , method:str = 'min' , bDA:bool = False ) -> list :
+    def func( r:float , c:float , lab:str='min' ) -> float :
+        if lab == 'max' :
+            return ( r if r > c else c )
+        if lab == 'min' :
+            return ( r if r < c else c )
     #
-    # CALUCULATES WHEN SAIGAS ARE LINKED
-    # MAX CORRESPONDS TO COMPLETE LINKAGE
-    # MIN CORRESPONDS TO SINGLE LINKAGE
+    nmind = np.argmin(D) # SIMPLE TIEBREAKER
+    if bDA :
+        planar_crds = lambda linear_crd,N : tuple( (int(linear_crd/N) , linear_crd%N) )
+        #
+        # HEURISTIC TIEBREAKER
+        dr    = D.reshape(-1)
+        ties  = np.where(dr==dr[nmind])[0]
+        ties  = ties[:int(len(ties)/2)]
+        if len(ties) > 1 :
+            nmind = ties[np.argmin( [ np.sum( D[planar_crds(t,len(D))[0],:]) for t in ties ]) ]
+    ( i,j )  = ( int(nmind/len(D)) , nmind%len(D) )
+    k = j - int(i<j)
+    l = i - int(j<i)
+    pop1 = nppop(D,i,j)
+    pop2 = nppop(pop1[-1],k,l)
+    lpr  = list(pop2[0])
+    d    = lpr.pop(l)
+    lpr  = np.array(lpr)
+    lpc  = pop2[1]
+    nvec = np.array([*[D[0,0]],*[ func(r,c,method) for (r,c) in zip(lpr,lpc) ]])
+    DUM  = np.eye(len(nvec))*0
+    DUM[ 0  , : ] = nvec
+    DUM[ :  , 0 ] = nvec
+    DUM[ 1: , 1:] = pop2[-1]
+    return ( [ DUM , (i,j) , d ]  )
+
+def linkage_dict_tuples ( D:np.array , method:str = 'min' ) -> dict :
+    N   = len(D)
+    dm  = np.max(D)*1.1
+    idx = list()
+    for i in range(N): D[i,i] = dm; idx.append(i)
+    cidx     = []
+    sidx     = set()
+    res      = [D]
+    linkages = dict()
+    while ( len(res[0]) > 1 ) :
+        res   = link1_ ( res[0] , method )
+        oidx  = tuple ( unpack( tuple( [ idx[i] for i in res[1] ]) ) )
+        unique_local_clusters = [ c for c in cidx if len( set(c) - set(oidx) ) >0 ]
+        unique_local_clusters .append( oidx )
+        cidx  .append( oidx )
+        sidx  = sidx|set(oidx)
+        idx   = [*unique_local_clusters[::-1] , *[i for i in range(N) if not i in sidx ]]
+        linkages[ oidx ] = res[-1]
+    for i in range(N) :
+        linkages[ (i,) ] = 0
+    return ( linkages )
+
+def link0_ ( D:np.array , method:str = 'min' ) -> list :
+    def func( r:float , c:float , lab:str='min' ) -> float :
+        if lab == 'max' :
+            return ( r if r > c else c )
+        if lab == 'min' :
+            return ( r if r < c else c )
+
+    nmind = np.argmin(D) # SIMPLE TIEBREAKER
+    ( i,j )  = ( int(nmind/len(D)) , nmind%len(D) )
+    k = j - int(i<j)
+    l = i - int(j<i)
+
+    pop1 = nppop(D,i,j)
+    pop2 = nppop(pop1[-1],k,l)
+    lpr  = list(pop2[0])
+    d    = lpr.pop(l)
+    lpr  = np.array(lpr)
+    lpc  = pop2[1]
+    nvec = np.array([*[D[0,0]],*[ func(r,c,method) for (r,c) in zip(lpr,lpc) ]])
+    DUM  = np.eye(len(nvec))*0
+    DUM[ 0  , : ] = nvec
+    DUM[ :  , 0 ] = nvec
+    DUM[ 1: , 1:] = pop2[-1]
+    return ( [ DUM , (i,j) , d ]  )
+
+def linkages_tiers ( D:np.array , method:str = 'min' ) -> dict :
+    N   = len(D)
+    dm  = np.max(D)*1.1
+    idx = list()
+    for i in range(N):  D[i,i] = dm ; idx.append( tuple((i,)) )
+    cidx     = []
+    sidx     = set()
+    res      = [D]
+    linkages = dict()
+    while ( len(res[0]) > 1 ) :
+        res          = link0_ ( res[0] , method )
+        found_cidx   = tuple( [ idx[i] for i in res[1] ])
+        idx = [ *[found_cidx], *[ix_ for ix_ in idx if not ix_ in set(found_cidx) ] ]
+        linkages[ found_cidx ] = res[-1]
+    for i in range(N) :
+        linkages[ (i,) ] = 0
+        D[i,i] = 0
+    return ( linkages )
+
+def lint2lstr ( seq:list[int] ) -> list[str] :
     #
-    D = distm
-    N = len(D)
+    # DUPLICATED IN special.lint2lstr
+    if isinstance ( seq,(list,tuple,set)) :
+        yield from ( str(x) for y in seq for x in lint2lstr(y) )
+    else :
+        yield seq
 
-    sidx = [ str(i)+'-'+str(j) for i in range(N) for j in range(N) if i<j ]
-    pidx = [ [i,j] for i in range(N) for j in range(N) if i<j ]
-    R = [ D[idx[0]][idx[1]] for idx in pidx ]
+def sclinkages ( distm ,command='min' , bStrKeys=True ) -> dict :
+    #
+    # FROM impetuous.clustering
+    # RETURN TYPE DIFFERS
+    #
+    from scipy.spatial.distance  import squareform
+    from scipy.cluster.hierarchy import linkage as sclinks
+    from scipy.cluster.hierarchy import fcluster
+    Z  = sclinks( squareform(distm) , {'min':'single','max':'complete'}[command] )
+    CL = {}
+    F  = {} # NEW
+    for d in Z[:,2] :
+        row  = fcluster ( Z ,d, 'distance' )
+        F[d] = row # NEW
+        sv_ = sorted(list(set(row)))
+        cl  = {s:[] for s in sv_}
+        for i in range( len( row ) ) :
+            cl[row[i]].append(i)
+        for v_ in list( cl.values() ) :
+            if tuple(v_) not in CL:
+                CL[tuple(v_)] = d
+    if bStrKeys :
+        L = {}
+        for item in CL.items():
+            L['.'.join( lint2lstr(item[0])  )] = item[1]
+        CL = L
+    return ( {'CL':CL , 'F':F} )
 
-    label_order = lambda i,k: i+'-'+k if '.' in i else k+'-'+i if '.' in k else  i+'-'+k if int(i)<int(k) else k+'-'+i
 
-    if command == 'max':
-        func  = lambda R,i,j : [(R[i],i),(R[j],j)] if R[i]>R[j] else [(R[j],j),(R[i],i)]
-    if command == 'min':
-        func  = lambda R,i,j : [(R[i],i),(R[j],j)] if R[i]<R[j] else [(R[j],j),(R[i],i)]
+def cluster_connections (       distm:np.array  , d:np.array ,
+                                command = 'min' , Z:np.array=None ,
+                                bReturnZ = False ) -> np.array :
+    #
+    from scipy.spatial.distance  import squareform
+    from scipy.cluster.hierarchy import linkage as sclinks
+    from scipy.cluster.hierarchy import fcluster
+    if Z is None :
+        Z       = sclinks( squareform(distm) , {'min':'single','max':'complete'}[command] )
+    row         = fcluster ( Z ,d, 'distance' )
+    if bReturnZ :
+        return ( Z )
+    return ( row )
 
-    LINKS = {}
-    cleared = set()
 
-    for I in range( N**2 ) : # WORST CASE SCENARIO
+def nearest_neighbor_graph_matrix ( distm:np.array, nn:int , bCheckSolution=False )->np.array:
+    desc__ = """
+     SLOW METHOD FOR CONSTRUCTING A NEAREST NEIGHBOR GRAPH
+     REPRESENTATION OF A DISTANCE MATRIX GIVEN THAT EACH
+     NODE SHOULD ONLY BE CONNECTED TO N NEIHGBOURS
+     A ROW CORRESPONDS TO A SPECIFIC NODES NEIGHBOURS
+     NOTE : THIS CREATES A NON-SYMMETRIC DISTANCE MATRIX
+    """
+    from scipy.spatial.distance import squareform
+    if len(np.shape(distm)) == 1 :
+        distm = squareform ( distm )
+    nn_distm = []
+    global_cval = -np.inf
+    for row in distm :
+        cval = sorted(row)[nn-1]
+        if cval > global_cval :
+            global_cval = cval
+        nrow = np.array([ rval if rval<=cval else np.inf for rval in row ])
+        nn_distm.append( nrow )
+    if bCheckSolution :
+        print (desc__ , '\n' , np.sum(np.array(nn_distm)<np.inf,1)  )
+    return ( np.array(nn_distm), global_cval )
 
-        clear = []
-        if len(R) == 0 :
-            break
 
-        nar   = np.argmin( R )
-        clu_idx = sidx[nar].replace('-','.')
+def CCA_DBSCAN ( eps:float, A:np.array , min_samples=1 ) -> np.array :
+    # SKLEARN DBSCAN WORKS ON COORDINATES NOT THE DISTANCE MATRIX AS INPUT
+    from sklearn.cluster import DBSCAN
+    clustering = DBSCAN( eps=eps , min_samples=min_samples ).fit(A)
+    return ( clustering.labels_ )
 
-        LINKS = { **{ clu_idx : R[nar] } , **LINKS }
-        lp    = {}
+def scipylinkages ( distm , command='min' , bStrKeys=True ) -> dict :
+    from scipy.spatial.distance  import squareform
+    from scipy.cluster.hierarchy import linkage as sclinks
+    from scipy.cluster.hierarchy import fcluster
+    Z = sclinks( squareform(distm) , {'min':'single','max':'complete'}[command] )
+    CL = {}
+    for d in Z[:,2] :
+        row = fcluster ( Z ,d, 'distance' )
+        sv_ = sorted(list(set(row)))
+        cl  = {s:[] for s in sv_}
+        for i in range( len( row ) ) :
+            cl[row[i]].append(i)
+        for v_ in list( cl.values() ) :
+            if tuple(v_) not in CL:
+                CL[tuple(v_)] = d
+    if bStrKeys :
+        L = {}
+        for item in CL.items():
+            L['.'.join( lint2lstr(item[0])  )] = item[1]
+        CL = L
+    return ( CL )
 
-        for l in range(len(sidx)) :
-            lidx = sidx[l]
-            lp [ lidx ] = l
+def linkages ( distm:np.array , command:str='min' ,
+               bStrKeys:bool = True , bUseScipy:bool = False ,
+               bMemSec=True, bLegacy:bool=False ) -> dict :
+    distm = np.array(distm)
+    if bMemSec :
+        distm = distm.copy()
+    if bUseScipy :
+        linkages_ = scipylinkages ( distm ,command=command , bStrKeys = False )
+    else :
+        linkages_ = linkages_tiers ( D = distm , method = command )
+    if bLegacy :
+        return ( linkage( distm = distm , command = command ) )
+    if bStrKeys :
+        L = {}
+        for item in linkages_.items():
+            L['.'.join( lint2lstr(item[0])  )] = item[1]
+        linkages_ = L
+    return ( linkages_ )
 
-        pij   = sidx[nar].split('-')
-        cidx  = set( unpack( [ s.split('-') for s in [ s for s in sidx if len(set(s.split('-'))-set(pij)) == 1 ] ] ) )-set(pij)
-        ccidx = set( [ s for s in [ s for s in sidx if len(set(s.split('-'))-set(pij)) == 1 ] ] )
-        found = {}
 
-        i   = pij[0]
-        j   = pij[1]
-
-        for k in cidx :
-            h0 , h1 , h2 , q0, J = None , None , None , None, 0
-            if k == j or k == i :
-                continue
-            la1 = label_order(i,k)
-            if la1 in lp :
-                J+=1
-                h1 = lp[ label_order(i,k) ]
-                h0 = h1 ; q0 = i
-            la2 = label_order(j,k)
-            if la2 in lp :
-                J+=1
-                h2 = lp[ label_order(j,k) ]
-                h0 = h2 ; q0 = j
-            if J == 2 :
-                Dijk = func ( R , h1 , h2 )
-            elif J == 1 :
-                Dijk = [[ R[h0],h0 ]]
-            else :
-                continue
-            nidx = [ s for s in sidx[Dijk[0][1]].split('-') ]
-            nidx = list( set(sidx[Dijk[0][1]].split('-'))-set(pij) )[0]
-            nclu_idx = clu_idx + '-' + nidx
-            found[ nclu_idx ] = Dijk[0][0]
-
-        clear = [*[ lp[c] for c in ccidx ],*[nar]]
-        cleared = cleared | set( unpack( [ sidx[c].split('-') for c in clear ]) )
-        R = rem(R,clear)
-        sidx = rem(sidx,clear)
-
-        for label,d in found.items() :
-            R.append(d)
-            sidx.append(label)
-    return ( LINKS )
-
-if __name__=='__main__' :
+if __name__ == '__main__' :
 
     D = [[0,9,3,6,11],[9,0,7,5,10],[3,7,0,9,2],[6,5,9,0,8],[11,10,2,8,0] ]
     print ( np.array(D) )
     print ( linkage( D, command='min') )
     print ( linkage( D, command='max') )
-    
+
     if False :
         #
         # TEST DEPENDS ON THE DIABETES DATA FROM BROAD INSTITUTE
@@ -827,10 +1215,9 @@ if __name__=='__main__' :
                         [9.00, 9.00, 9.00, 0.00, 0.10, 0.10],
                         [9.10, 9.00, 9.00, 0.10, 0.00, 0.15],
                         [9.10, 9.00, 9.00, 0.10, 0.15, 0.00] ] )
-        print ( connectivity(A,0.11) ) # 
+        print ( connectivity(A,0.11) )
         print ( dbscan(distance_matrix=pd.DataFrame(A).values,eps=0.11,minPts=2) )
 
-
     import time
     print ( 'HERE' )
 
