@@ -173,7 +173,6 @@ def create_encoding_data_frame ( journal_df , formula , bVerbose = False ) :
                             journal_df.loc[ [ c.replace(' ','') for c in formula.split('~')[1].split('+') if not 'C(' in c] , : ] ]).T
     return ( encoding_df.apply(pd.to_numeric) )
 
-
 def interpret_problem ( analyte_df , journal_df , formula , bVerbose=False ) :
     #
     # THE JOURNAL_DF IS THE COARSE GRAINED DATA (THE MODEL)
@@ -271,6 +270,8 @@ def calculate_alignment_properties ( encoding_df , quantx, quanty, scorex,
                   index = use_labels , columns=analyte_df.index.values )
     #
     # print ( 'P VALUES ALIGNED TO PLS AXES' )
+    #print ( proj_df )
+    #exit(1)
     for idx in proj_df.index :
         proj_p,proj_rho = quantify_density_probability ( proj_df.loc[idx,:].values )
         proj_df = proj_df.rename( index = {idx:idx+',r'} )
@@ -386,10 +387,11 @@ def run_shape_alignment_clustering ( analyte_df , journal_df , formula, bVerbose
         return ( res_df , clusters_df )
 
 
-def kmeans_clustering_alignment ( P, Q , bHighDim = False ) :
-    return ( knn_clustering_alignment( P , Q , bHighDim = bHighDim ) )
+def knn_clustering_alignment ( P, Q , bHighDim = False ) :
+    print ( "DOING KMEANS ALIGNMENT INSTEAD" )
+    return ( kmeans_clustering_alignment( P , Q , bHighDim = bHighDim ) )
 
-def knn_clustering_alignment( P , Q , bHighDim=False ) : # ITS A SEEDED KMEANS THE NAME KNN IS A MISNOMER
+def kmeans_clustering_alignment( P , Q , bHighDim=False ) :
 
     NOTE_ = "This is just a standard kmeans in arbitrary dimensions that start out with centroids that have been shape aligned"
     ispanda = lambda P: 'pandas' in str(type(P)).lower()
@@ -459,7 +461,7 @@ def multifactor_evaluation (  analyte_df , journal_df , formula ) :
     for c in eval_df.columns :
         all.append ( pd.DataFrame ( quantify_density_probability ( eval_df.loc[:,c].values ),
                 index = [c+',p',c+',r'], columns=eval_df.index ).T)
-    res_df = pd.concat( all,1 )
+    res_df = pd.concat( all,axis=1 )
     for c in res_df.columns:
         if ',p' in c:
             q = [ qv[0] for qv in qvalues(res_df.loc[:,c].values) ]
@@ -510,8 +512,6 @@ def regression_assessment ( model , X , y , bLog = False ) :
     #
     return ( mstat )
 
-
-
 def proj_c ( P ) :
     # P CONTAINS MUTUTALLY ORTHOGONAL COMPONENTS ALONG THE COLUMNS
     # THE CS CALCULATION MIGHT SEEM STRANGE BUT FULLFILS THE PURPOSE
@@ -519,11 +519,10 @@ def proj_c ( P ) :
         print ( "FUNCTION REQUIRES A SAIGA OR PANDA DATA FRAME" )
     CS  = P.T.apply( lambda x: pd.Series( [x[0],x[1]]/np.sqrt(np.sum(x**2)),index=['cos','sin']) ).T
     RHO = P.T.apply( lambda x: np.sqrt(np.sum(x**2)) )
-    CYL = pd.concat( [RHO*CS['cos'],RHO*CS['sin']],1 )
+    CYL = pd.concat( [RHO*CS['cos'],RHO*CS['sin']],axis=1 )
     CYL.columns = ['X','Y']
     return ( CYL )
 
-
 def multivariate_factorisation ( analyte_df , journal_df , formula ,
                           bVerbose = False , synonyms = None , blur_cutoff = 99.8 ,
                           exclude_labels_from_centroids = [''] ,
@@ -588,6 +587,23 @@ def associations ( M , W = None , bRanked = True ) :
     res = PQ/R2
     return ( res )
 
+def spearmanrho ( xs , ys ) :
+    from scipy.stats import rankdata
+    xs_ = np.array( [ rankdata(x,'average') for x in xs] )
+    ys_ = np.array( [ rankdata(y,'average') for y in ys] )
+    return ( correlation_core(xs_,ys_) )
+
+def pearsonrho ( xs , ys ) :
+    return ( correlation_core ( xs , ys ) )
+
+def correlation_core ( xs , ys ) :
+    x_means = xs.mean(axis=1)
+    y_means = ys.mean(axis=1)
+    xms = xs - (x_means*np.array([[1 for i in range(np.shape(xs)[1]) ]]).T).T # THIS CAN BE IMPROVED
+    yms = ys - (y_means*np.array([[1 for i in range(np.shape(ys)[1]) ]]).T).T # THIS CAN BE IMPROVED
+    r = np.dot( yms , xms.T ) / np.sqrt( (yms*yms).sum(axis=1).reshape(len(yms),1) @ (xms*xms).sum(axis=1).reshape(1,len(xms))  )
+    return ( r )
+
 
 crop = lambda x,W:x[:,:W]
 def run_shape_alignment_regression( analyte_df , journal_df , formula ,
@@ -694,7 +710,6 @@ def qvalues ( p_values_in , pi0 = None ) :
     return qs_
 
 
-
 class Qvalues ( object ) :
     def __init__( self, pvalues:np.array , method:str = "UNIQUE" , pi0:np.array = None ) :
         from scipy.stats import rankdata
@@ -764,6 +779,113 @@ class Qvalues ( object ) :
         self.qvalues = np.array([q[0] for q in qs_])
         return np.array(qs_)
 
+class Pvalues ( object ) :
+    def __init__( self, data_values:np.array , method:str = "RANK DERIV E" ) :
+        from scipy.stats import rankdata
+        self.rankdata = rankdata
+        self.method   : str        = method
+        self.dvalues  : np.array   = data_values
+        self.pvalues  : np.array   = None
+        self.dsdrvalues : np.array = None
+        self.dpres    : np.array   = None
+        if method == "RANK DERIV E" :
+            self.dpres = self.pvalues_dsdr_e ( self.dvalues , True)
+        if method == "RANK DERIV N" :
+            self.dpres = self.pvalues_dsdr_n ( self.dvalues , True )
+        if method == "NORMAL" :
+            self.dpres = self.normal_pvalues ( self.dvalues , True )
+        self.pvalues = self.dpres[0]
+
+    def __str__ ( self ) :
+        return ( self.info() )
+
+    def __repr__( self ) :
+        return ( self.info() )
+
+    def help ( self ) :
+        #
+        # PVALUES FROM "RANK DERIVATIVES"
+        #
+        desc__ = "\n\nRANK DERIVATIVE P-VALUES\nVIABLE METHODS ARE method = RANK DERIV E, RANK DERIV N \n\n EMPLOYED METHOD: " + self.method
+        return ( desc__ )
+
+    def info ( self ) :
+        desc__ = "\nMETHOD:"+self.method+"\n   p-values       \t     ds-values\n"
+        return ( desc__+'\n'.join( [ ' \t '.join(["%10.10e"%z for z in s]) for s in self.dpres.T ] ) )
+
+    def get ( self ) :
+        return ( self.qpres )
+
+    def sgn ( self, x:float) -> int :
+        return( - int(x<0) + int(x>=0) )
+
+    def nn ( self, N:int , i:int , n:int=1 )->list :
+        t = [(i-n)%N,(i+n)%N]
+        if i-n<0 :
+            t[0] = 0
+            t[1] += n-i
+        if i+n>=N :
+            t[0] -= n+i-N
+            t[1] = N-1
+        return ( t )
+
+    def normal_pvalues ( self, v:np.array , bReturnDerivatives:bool=False  ) -> np.array :
+        ds = v # TRY TO ACT LIKE YOU ARE NORMAL ...
+        N = len(v)
+        M_ , Var_ = np.mean(ds) , np.std(ds)**2
+        from scipy.special import erf as erf_
+        loc_Q   = lambda X,mean,variance : [ 1. - 0.5*( 1. + erf_(  (x-mean)/np.sqrt( 2.*variance ) ) ) for x in X ]
+        rv = loc_Q ( ds,M_,Var_ )
+        if bReturnDerivatives :
+            rv = [*rv,*ds ]
+        return ( np.array(rv).reshape(-1,N) )
+
+    def pvalues_dsdr_n ( self, v:np.array ,
+                         bReturnDerivatives:bool=False ,
+                         bSymmetric:bool=True ) -> np.array :
+        #
+        N = len(v)
+        vsym = lambda a,b : a*self.sgn(a) if b else a
+        import scipy.stats as st
+        rv = st.rankdata(v,'ordinal') - 1
+        vr = { int(k):v for k,v in zip(rv,range(len(rv)))}
+        ds = []
+        for w,r in zip(v,rv) :
+            nr  = self.nn(N,int(r),1)
+            nv  = [ vr[j] for j in nr ]
+            s_  = [ v[j] for j in sorted(list(set( [ *[vr[int(r)]] , *nv ] )) ) ]
+            dsv = np.mean( np.diff(s_) )
+            ds.append( vsym( dsv , bSymmetric) ) # DR IS ALWAYS 1
+        M_,Var_ = np.mean(ds) , np.std(ds)**2
+        from scipy.special import erf as erf_
+        loc_Q   = lambda X,mean,variance : [ 1. - 0.5*( 1. + erf_(  (x-mean)/np.sqrt( 2.*variance ) ) ) for x in X ]
+        rv = loc_Q ( ds,M_,Var_ )
+        if bReturnDerivatives :
+            rv = [*rv,*ds ]
+        return ( np.array(rv).reshape(-1,N) )
+
+    def pvalues_dsdr_e ( self, v:np.array ,
+                         bReturnDerivatives:bool=False ,
+                         bSymmetric:bool=True ) -> np.array :
+        #
+        N = len(v)
+        vsym = lambda a,b : a*self.sgn(a) if b else a
+        import scipy.stats as st
+        rv = st.rankdata(v,'ordinal') - 1
+        vr = { int(k):v for k,v in zip(rv,range(len(rv)))}
+        ds = []
+        for w,r in zip(v,rv) :
+            nr  = self.nn(N,int(r),1)
+            nv  = [ vr[j] for j in nr ]
+            s_  = [ v[j] for j in sorted(list(set( [ *[vr[int(r)]] , *nv ] )) ) ]
+            dsv = np.mean( np.diff(s_) )
+            ds.append( vsym( dsv , bSymmetric) ) # DR IS ALWAYS 1
+        M_ = np.mean ( ds )
+        loc_E  = lambda X,L_mle : [ np.exp(-L_mle*x) for x in X ]
+        ev = loc_E ( ds,1.0/M_)   # EXP DISTRIBUTION P
+        if bReturnDerivatives :
+            rv = [*ev,*ds ]
+        return ( np.array(rv).reshape(-1,N) )
 
 
 class MultiFactorAnalysis ( object ) :
@@ -1116,19 +1238,6 @@ def mycov( x , full_matrices=0 ):
     return C / (x.shape[0]-1)
 
 from scipy.special import chdtrc as chi2_cdf
-def p_value_merger ( pvalues_df , p_label=',p' , axis = 0 ) :
-    #
-    print( " REQUIRED READING: doi: 10.1093/bioinformatics/btw438" )
-    print( " ALSO MAKE SURE TO ADD THAT ARTICLE AS ADDITIONAL CITATION" )
-    print( " IF THIS METHOD IS EMPLOYED" )
-    print( " READ ABOVE ! " )
-    print( " YOU CALLED A FUNCTION FOR MERGING P-VALUES" )
-    print( " THIS METHOD IS NO LONGER SUPPORTED " )
-    print( " FOR IMPETUOUS VERSIONS > 0.84.0 " )
-    print( " THE METHOD merge_significance COULD BE EMPLOYED INSTEAD ")
-    print( " BUT IT IS NOT RECOMMENDED " )
-    print( " FATAL : WILL TERMINATE NOW " )
-    exit(1)
 
 def parse_test ( statistical_formula, group_expression_df , journal_df , test_type = 'random' ) :
     #
@@ -1182,24 +1291,6 @@ def prune_journal ( journal_df , remove_units_on = '_' ) :
     journal_df = pd.concat( [nmr_journal,str_journal] )
     return( journal_df )
 
-def merge_significance ( significance_df , distance_type='euclidean' ) :
-    # TAKES P VALUES OR Q VALUES
-    # TRANSFORMS INTO A MERGED P OR Q VALUE VIA
-    # THE DISTANCE SCORE
-    # THE DATA DOMAIN SIGNIFICANCE IS ALONG COLUMNS AND
-    # GROUPS ALONG INDICES
-    # EX: pd.DataFrame( np.random.rand(20).reshape(5,4) , columns=['bio','cars','oil','money']).apply( lambda x: -1.*np.log10(x) ).T.apply( lambda x: np.sqrt(np.sum(x**2)) )
-    #
-    distance = lambda x : np.sqrt(np.sum(x**2))
-    if distance_type == 'euclidean' : # ESTIMATE
-        distance = lambda x : np.sqrt(np.sum(x**2))
-    if distance_type == 'extreme' :   # ANTI-CONSERVATIVE ESTIMATE
-        distance = lambda x : np.max(x)
-    if distance_type == 'mean' :      # MEAN ESTIMATE
-        distance = lambda x : np.mean(x)
-    get_pvalue = lambda x : 10**(-x)
-    return ( significance_df.apply( lambda x: -1.*np.log10(x) ).T.apply(distance).apply(get_pvalue) )
-
 def group_significance( subset , all_analytes_df = None ,
                         tolerance = 0.05 , significance_name = 'pVal' ,
                         AllAnalytes = None , SigAnalytes = None,
@@ -1262,10 +1353,6 @@ def quantify_groups_by_analyte_pvalues( analyte_df, grouping_file, delimiter='\t
     return ( edf.T )
 
 class APCA ( object ) :
-    #
-    # THIS CLASS PERFORMS A SPARSE PCA IF REQUESTED
-    # IT THEN USES THE SPARSE SVD ALGORITHM FOUND IN SCIPY
-    # THE STANDARD IS TO USE THE NUMPY SVD
     #
     def __init__ ( self , X = None , k =-1 , fillna = None , transcending = True , not_sparse = True ) :
         from scipy.sparse import csc_matrix
@@ -1377,7 +1464,7 @@ def quantify_groups ( analyte_df , journal_df , formula , grouping_file , synony
             edf.loc[l] = q
     return ( edf.T )
 
-from scipy.stats import combine_pvalues
+from scipy.stats import combine_pvalues # IS THIS USED...
 
 def quantify_by_dictionary ( analyte_df , journal_df , formula , split_id=None,
                     grouping_dictionary = dict() , synonyms = None ,
@@ -1569,6 +1656,74 @@ def differential_analytes ( analyte_df , cols = [['a'],['b']] ):
     ddf = ddf.sort_values('Dist', ascending = False )
     return ( ddf )
 
+def single_fc_compare ( df:pd.DataFrame, what:str, levels:list[str] = None, bVerbose:str=True ,
+                    sep:str=', ' , bRanked:bool = False , bLogFC:bool=False ) :
+
+        from scipy.stats import mannwhitneyu
+        from scipy.stats import ttest_ind
+
+        if levels is None :
+            levels = df .loc[ what,: ].values
+
+        if len ( levels ) == 1 :
+            levels .append( 'Background' )
+            df .loc[what] = [ levels[0] if levels[0] in v else 'Background' for v in df.loc[what].values ]
+            if bVerbose:
+                print ( df.loc[what] )
+                print ( 'HERE' , levels )
+                from collections import Counter
+                print ( Counter( df.loc[what].values.tolist() ) )
+
+        ddf = df.iloc[[ not what in i for i in df.index.values],:]
+        C0 = []
+        C1 = []
+        C2 = []
+        C3 = []
+
+        level1 = levels[0]
+        level2 = levels[1]
+
+        compared = str(level1)+sep+str(level2)
+        C3.append(compared)
+
+        X = []
+        Y = []
+        Z = []
+        df1 = ddf.iloc[ :, [level1 == v for v in df.loc[what,:].values] ]
+        df2 = ddf.iloc[ :, [level2 == v for v in df.loc[what,:].values] ]
+
+        if level1 == level2 :
+            for i_ in range( len(ddf) ) :
+                X.append( 0 )
+                Y.append( 1 )
+                Z.append( 0 )
+        else :
+            for i_ in range(len(ddf)):
+                v = df1.iloc[i_,:].values
+                w = df2.iloc[i_,:].values
+                if bRanked :
+                    stats = mannwhitneyu( x=v.tolist() , y=w.tolist() )
+                    if bLogFC : # WARNING SHOULD NOT BE FORCED
+                        f = np.log2( np.mean(v)+1 ) - np.log2( np.mean(w)+1 )
+                    else :
+                        f = np.median(v) - np.median(w)
+                else :
+                    stats = ttest_ind(  v.tolist() , w.tolist()  )
+                    if bLogFC : # WARNING SHOULD NOT BE FORCED
+                        f = np.log2( np.mean(v)+1 )-np.log2( np.mean(w)+1 )
+                    else :
+                        f = np.mean( v ) - np.mean( w )
+                s = stats[0]
+                p = stats[1]
+                X.append(s)
+                Y.append(p)
+                Z.append(f)
+        C0.append(X)
+        C1.append(Y)
+        C2.append(Z)
+        return ( {'statistic':C0, 'p-value':C1, 'contrast':C2, 'comparison':C3, 'index':ddf.index.values} )
+
+
 def add_kendalltau( analyte_results_df , journal_df , what='M' , sample_names = None, ForceSpearman=False ) :
     # ADD IN CONCORDANCE WITH KENDALL TAU
     if what in set(journal_df.index.values) :
@@ -1617,6 +1772,247 @@ def quality_metrics ( TP:int , FP:int , FN:int , TN:int , alternative:str='two-s
         }
     return ( results_lookup )
 
+def invert_dict ( dictionary ) :
+    inv_dictionary = dict()
+    if True :
+        for item in dictionary.items() :
+            if item[1] in inv_dictionary :
+                inv_dictionary[item[1]] .append(item[0])
+            else :
+                inv_dictionary[item[1]] = [item[0]]
+    return ( inv_dictionary )
+
+
+def correlation_distance ( agg_df:pd.DataFrame  , decimal_power:int = 10000 ,
+                           bLegacy:bool = False , correlation_type:str='spearman',
+                           correlation_transform = lambda x:x ) :
+    if bLegacy :
+        import scipy.stats as ss
+        spearman_df = ss.spearmanr( agg_df )[0]
+        distm       = pd.DataFrame( np.sqrt( 1 - spearman_df ) ,
+                                    index   = agg_df.columns.values ,
+                                    columns = agg_df.columns.values )
+        distm = distm.apply(lambda x: np.round(x*decimal_power)/decimal_power )
+    else :
+        if correlation_type == 'pearson' :
+            corr =  pearsonrho(agg_df.values,agg_df.values)
+        else :
+            corr = spearmanrho(agg_df.values,agg_df.values)
+
+        distm = pd.DataFrame( np.sqrt( 1 - correlation_transform(corr) ) ,
+                                    index   = agg_df.columns.values ,
+                                    columns = agg_df.columns.values )
+        distm = distm.apply(lambda x: np.round(x*decimal_power)/decimal_power )
+    return ( distm )
+
+def confusion_matrix_df ( dict_row:dict , dict_col:dict , bSwitchKeyValues:str=True, bCheck:str=False ) -> pd.DataFrame :
+    if bSwitchKeyValues :
+        dict_row = invert_dict(dict_row)
+        dict_col = invert_dict(dict_col)
+    all_interactions = list(dict_row.keys())
+    if bCheck:
+        all_interactions = sorted(list(set( dict_row.keys() ) | set( dict_col.keys() )))
+    num_p = len(all_interactions)
+    confusion = np.zeros(num_p*num_p).reshape(num_p,num_p)
+    for i in range(num_p) :
+        for j in range(num_p) :
+            confusion[i,j] = len( set(dict_row[all_interactions[i]]) & set(dict_col[all_interactions[j]]) )
+    return ( pd.DataFrame(confusion,columns=all_interactions,index=all_interactions) )
+
+def scaler_transform( df,i=0 ) :
+            from impetuous.special import zvals
+            z_analytes_df = df
+            if (i+1)%2 == 0 :
+                z_analytes_df = z_analytes_df.T
+            z_analytes_df = pd.DataFrame( zvals(z_analytes_df.values)['z'] ,
+                         columns = z_analytes_df.columns, index=z_analytes_df.index )
+            z_analytes_df = z_analytes_df[~z_analytes_df.isin([np.inf,np.nan,-np.inf]).any(1)]
+            if (i+1)%2 == 0 :
+                z_analytes_df = z_analytes_df.T
+            return ( z_analytes_df )
+
+def local_pca ( df, ndims = None  ) :
+    from sklearn.decomposition import PCA
+    pca     = PCA ( ndims )
+    scores  = pca.fit_transform( df.values )
+    weights = pca.components_.T
+    return ( scores , weights , df.index, df.columns )
+
+def compositional_analysis(x:np.array)->list[float]:
+    # https://doi.org/10.1093/bib/bbw008
+    # Tau, Gini, TSI, SPM
+    n           = len(x)
+    tau         = np.sum(1-x/np.max(x))/(n-1)
+    gini        = np.sum( np.array([np.abs(xi-xj) for xi in x for xj in x])/2/n**2/np.mean(x) )
+    geni        = gini*n/(n-1)
+    TSI         = np.max(x)/np.sum(x)
+    beta        = np.sum(1-x/np.max(x))/n # own version, works for single component compositions
+    return ( [beta , tau , gini , geni, TSI, (n-1)/n ] )
+
+
+def multivariate_aligned_pca ( analytes_df , journal_df ,
+                sample_label = 'Sample ID', align_to = 'Modulating group' , n_components=None ,
+                add_labels = ['Additional information'] , e2s=None , color_lookup=None , ispec=None ) :
+    # SAIGA PROJECTIONS
+    what                = align_to
+    analytes_df         = analytes_df.loc[:,journal_df.columns.values]
+    dict_actual         = { sa:di for sa,di in zip(*journal_df.loc[[sample_label,what],:].values) }
+    sample_infos        = []
+    if not add_labels is None :
+        for label in add_labels :
+            sample_infos.append( tuple( (label, { sa:cl for cl,sa in zip(*journal_df.loc[[label,sample_label]].values) } ) ) )
+    #
+    N_P = n_components
+    if n_components is None:
+        N_P = np.min(np.shape(analytes_df))
+    scores,weights,nidx,ncol = local_pca( scaler_transform( analytes_df.copy() ) , ndims = N_P )
+    #
+    pcaw_df = pd.DataFrame(weights , columns=['PCA '+str(i+1) for i in range(N_P) ] , index=ncol )
+    pcas_df = pd.DataFrame( scores , columns=['PCA '+str(i+1) for i in range(N_P) ] , index=nidx )
+
+    from scipy.stats import rankdata
+    corr_r = rankdata(pcas_df.T.apply(lambda x:np.sum(x**2)).values)/len(pcas_df.index)
+
+    pcaw_df .loc[:,what] = [ dict_actual[s] for s in pcaw_df.index.values ]
+    projection_df       = pcaw_df.groupby(what).apply(np.mean)
+    projection_df       = ( projection_df.T / np.sqrt(np.sum(projection_df.values**2,1)) ).T
+    projected_df        = pd.DataFrame( np.dot(projection_df,pcas_df.T), index=projection_df.index, columns=pcas_df.index )
+    owners  = projected_df.index.values[projected_df.apply(np.abs).apply(np.argmax).values]
+    if ispec is None :
+        ispec = int( len(projection_df)>1 )
+    specificity = projected_df.apply(np.abs).apply(lambda x:compositional_analysis(x)[ispec] ).values
+
+    pcas_df .loc[:,'Owner'] = owners
+    pcaw_df = pcaw_df.rename( columns={what:'Owner'} )
+    pcas_df.loc[:,'Corr,r'] = corr_r
+    pcas_df.loc[:,'Spec,' + {0:'beta',1:'tau',2:'gini',3:'geni'}[ ispec ] ] = specificity
+
+    if not color_lookup is None :
+        pcas_df.loc[:, 'Color'] = [ color_lookup[o] for o in pcas_df.loc[:,'Owner'] ]
+        pcaw_df.loc[:, 'Color'] = [ color_lookup[o] for o in pcaw_df.loc[:,'Owner'] ]
+
+    if not e2s is None :
+        pcas_df.loc[:,'Symbol'] = [ (e2s[g] if not 'nan' in str(e2s[g]).lower() else g) if g in e2s else g for g in pcas_df.index.values ]
+    #
+    if len( sample_infos ) > 0 :
+        for label_tuple in sample_infos :
+            label = label_tuple[0]
+            sample_lookup = label_tuple[1]
+            if not sample_lookup is None :
+                pcaw_df.loc[ :, label ] = [ sample_lookup[s] for s in pcaw_df.index.values ]
+    return ( pcas_df , pcaw_df )
+
+def sort_matrix ( matrix:np.array , linkage:str='single' ) -> list[int] :
+    if linkage != 'dev' :
+        from scipy.cluster import hierarchy
+        from scipy.spatial.distance import squareform
+        distm   = absolute_coordinates_to_distance_matrix ( matrix )
+        pdi     = squareform ( distm )
+        Z       = hierarchy.linkage( pdi , linkage )
+        labels  = None
+        if labels is None :
+            labels = ['LID'+str(i) for i in range(len(distm)) ]
+        dn = hierarchy.dendrogram( Z ,labels=labels )
+        order = dn ['leaves']
+        return ( order )
+
+    if linkage == 'dev' :
+        # DEV
+        from scipy.stats import rankdata
+        order = rankdata( np.mean(matrix,1) ,'ordinal') - 1
+        return ( order )
+
+def jaccard ( A:set,B:set ) -> float :
+    return ( len( set(A)&set(B) )/ len( set(A)|set(B) ) )
+
+def jaccard_distance(A:set,B:set) -> float :
+    return ( ( len(set(A)|set(B)) - len(set(A)&set(B)) ) / len(set(A)|set(B)) )
+
+def blind_confusion ( v1:list[str] , v2:list[str] , sort_type:str=None , bAbsolute:bool = True ) -> list[np.array] :
+    CM,BM = [],[]
+    for i in range( len(v1) ) :
+        w,v = [],[]
+        for j in range( len(v2) ) :
+            w.append ( jaccard( set(v1[i]) , set(v2[j]) ) ) # WE DO DISTANCE SORTING LATER
+            v.append ( len(set(v1[i])&set(v2[j])) )
+        CM.append ( np.array(w) )
+        BM.append ( np.array(v) )
+    MAT = np.array( CM )
+    if bAbsolute :
+        QAT = np.array( BM )
+    if not sort_type is None :
+        something_ok = set(['single','dev','complete','ward','median'])
+        if not sort_type in something_ok :
+            print ( 'ERROR: SET sort_type TO' , something_ok , '\nSUGGEST USING: single' )
+        i_order = sort_matrix( MAT      ,       linkage = sort_type )
+        j_order = sort_matrix( MAT.T    ,       linkage = sort_type )
+        if bAbsolute :
+            QAT = np.array( [ QAT[ i,j ] for i in i_order for j in j_order ]).reshape( *np.shape(QAT) )
+        else :
+            MAT = np.array( [ MAT[ i,j ] for i in i_order for j in j_order ]).reshape( *np.shape(MAT) )
+    if bAbsolute :
+        return ( [ QAT,i_order,j_order ] )
+    else :
+        return ( [ MAT,i_order,j_order ] )
+
+def confusion_matrix ( dict_row:dict , dict_col:dict , bSwitchKeyValues:bool=True ) -> dict :
+    if bSwitchKeyValues :
+        dict_row = invert_dict(dict_row)
+        dict_col = invert_dict(dict_col)
+    all_interactions = list(dict_row.keys())
+    num_p = len(all_interactions)
+    confusion = np.zeros(num_p*num_p).reshape(num_p,num_p)
+    for i in range(num_p) :
+        for j in range(num_p) :
+            confusion[i,j] = len( set(dict_row[all_interactions[i]]) & set(dict_col[all_interactions[j]]) )
+    return ( {'confusion matrix':confusion,'index names':all_interactions } )
+
+def group_classifications ( df:pd.DataFrame     ,
+                det_limit:float         = 1.0   ,
+                log2FClim:float         = 2.0   ,
+                bLog2:bool              = True ) -> dict() :
+    #
+    first_quartile_rank  = lambda N:int(np.round( 1/4 * N ))
+    second_quartile_rank = lambda N:int(np.round( 1/2 * N ))
+    third_quartile_rank  = lambda N:int(np.round( 3/4 * N ))
+    #
+    n_  = len ( df.index  .values )
+    m_  = len ( df.columns.values )
+    agg_df = df
+    #
+    if ( bLog2 ) :
+        agg_df = df .apply( lambda x:np.log2(x+1) )
+    #
+    grp_type_0  = set( agg_df.index.values[ np.sum(agg_df,1) <= det_limit  ] )
+    grp_type_1  = set( agg_df.index.values[ np.sum(agg_df,1) >  det_limit  ] )
+    #
+    ran_df = agg_df.copy()
+    ran_df .columns = range(m_)
+    ran_df = ran_df.T.apply(sorted).T
+    #
+    max_n_group = m_ - third_quartile_rank( m_ )
+    grp_type_4  = set( ran_df.index.values[ ran_df.iloc[:,-1] - ran_df.iloc[:,-2] > log2FClim ])
+    grp_type_3  = set( ran_df.index.values[\
+        np.mean(ran_df.iloc[:,-max_n_group:],1) - np.max(ran_df.iloc[:,:-max_n_group],1) > log2FClim \
+        ]) - grp_type_4
+    #
+    max_n_group = second_quartile_rank(m_)
+    grp_type_2 = set( ran_df.index.values[\
+                    np.mean(ran_df.iloc[:,-max_n_group:],1) - np.mean(ran_df.iloc[:,:-max_n_group],1) > log2FClim \
+            ]) - (grp_type_3|grp_type_4)
+
+    grp_type_1  = grp_type_1 - (grp_type_4|grp_type_3|grp_type_2)
+    #
+    results_d =\
+    { '4' :  grp_type_4   ,
+      '3' :  grp_type_3   ,
+      '2' :  grp_type_2   ,
+      '1' :  grp_type_1   ,
+      '0' :  grp_type_0   }
+    return ( results_d )
+
+
+
 def pvalues_dsdr_n ( v:np.array , bReturnDerivatives:bool=False ) -> np.array :
     #
     # PVALUES FROM "RANK DERIVATIVES"
@@ -1634,14 +2030,14 @@ def pvalues_dsdr_n ( v:np.array , bReturnDerivatives:bool=False ) -> np.array :
         return ( t )
 
     import scipy.stats as st
-    rv = st.rankdata(v)-1
+    rv = st.rankdata(v,'ordinal')-1
     vr = { int(k):v for k,v in zip(rv,range(len(rv)))}
     ds = []
     for w,r in zip(v,rv) :
-        nr = nn(N,int(r),1)
-        nv = [ vr[j] for j in nr]
-        s_ = sorted(list(set([*[v[vr[int(r)]]], *[v[j] for j in nv]])) )
-        ds.append( np.mean( np.diff(s_) )) # DR IS ALWAYS 1
+        nr  = nn(N,int(r),1)
+        nv  = [ vr[j] for j in nr]
+        s_  = [ v[j] for j in sorted(list(set( [ *[vr[int(r)]] , *nv ] )) ) ]
+        ds.append( np.abs(np.mean(np.diff(s_))) ) # DR IS ALWAYS 1
     M_,Var_ = np.mean(ds) , np.std(ds)**2
     from scipy.special import erf as erf_
     loc_Q   = lambda X,mean,variance : [ 1. - 0.5*( 1. + erf_(  (x-mean)/np.sqrt( 2.*variance ) ) ) for x in X ]
@@ -1670,17 +2066,18 @@ def pvalues_dsdr_e ( v:np.array , bReturnDerivatives:bool=False ) -> np.array :
         return ( t )
 
     import scipy.stats as st
-    rv = st.rankdata(v)-1
+    rv = st.rankdata(v,'ordinal')-1
     vr = { int(k):v for k,v in zip(rv,range(len(rv)))}
     ds = []
     for w,r in zip(v,rv) :
-        nr = nn(N,int(r),1)
-        nv = [ vr[j] for j in nr]
-        s_ = sorted(list(set([*[v[vr[int(r)]]], *[v[j] for j in nv]])) )
-        ds.append( np.mean( np.diff(s_) )) # DR IS ALWAYS 1
+        nr  = nn(N,int(r),1)
+        nv  = [ vr[j] for j in nr]
+        s_  = [ v[j] for j in sorted(list(set( [ *[vr[int(r)]] , *nv ] )) ) ]
+        ds.append( np.abs(np.mean(np.diff(s_))) ) # DR IS ALWAYS 1
     M_ = np.mean(ds)
     loc_E  = lambda X,L_mle : [ np.exp(-L_mle*x) for x in X ]
     ev = loc_E ( ds,1.0/M_)   # EXP DISTRIBUTION P
+    rv = np.array([ev])
     if bReturnDerivatives :
         rv = [*ev,*ds ]
     return ( np.array(rv).reshape(-1,N) )
@@ -1735,7 +2132,6 @@ def calculate_rates( journal_df:pd.DataFrame , inferred_df:pd.DataFrame ,
 
     return ( results_lookup )
 
-
 def assign_quality_measures( journal_df , result_dfs ,
                              formula , inference_label='owner' ,
                              plabel = ',p' , qlabel = ',q' ) :
@@ -1775,6 +2171,7 @@ def fibonacci(n:int) -> int :
 def f_truth(i:int) -> bool : #  THE SQUARE SUM OF THE I:TH AND I+1:TH FIBONACCI NUMBER ARE EQUAL TO THE FIBONACCI NUMBER AT POSITION 2I+1
     return ( fibonacci(i)**2+fibonacci(i+1)**2 == fibonacci(2*i+1))
 
+
 if __name__ == '__main__' :
 
     test_type = 'random'
@@ -1789,7 +2186,6 @@ if __name__ == '__main__' :
 
     print ( quantify_groups( analyte_df, journal_df, 'Group ~ Var + C(Cat) ', grouping_file ) )
 
-
     import impetuous.quantification as impq
     import sys
     sys.setrecursionlimit(20000)
